#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <objbase.h>
#include <commctrl.h>
#include <richedit.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "StackFunc.h"
#include "ConvFunc.h"
#include "ISystemFunction.h"
#include "IDocument.h"
#include "IServer.h"
#include "Scripts.h"
#include "Resources\Resource.h"


//Global variables
ITypeInfo *g_SystemFunctionTypeInfo;

const ISystemFunctionVtbl MyISystemFunctionVtbl={
  SystemFunction_QueryInterface,
  SystemFunction_AddRef,
  SystemFunction_Release,
  SystemFunction_GetTypeInfoCount,
  SystemFunction_GetTypeInfo,
  SystemFunction_GetIDsOfNames,
  SystemFunction_Invoke,
  SystemFunction_AddParameter,
  SystemFunction_Call,
  SystemFunction_GetLastError
};


//// ISystemFunction

HRESULT STDMETHODCALLTYPE SystemFunction_QueryInterface(ISystemFunction *this, REFIID vTableGuid, void **ppv)
{
  if (!ppv) return E_POINTER;

  if (IsEqualIID(vTableGuid, &IID_IUnknown) || IsEqualIID(vTableGuid, &IID_IDispatch))
  {
    *ppv=this;
    this->lpVtbl->SystemFunction_AddRef(this);
    return NOERROR;
  }
  *ppv=NULL;
  return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE SystemFunction_AddRef(ISystemFunction *this)
{
  return ++((IRealSystemFunction *)this)->dwCount;
}

ULONG STDMETHODCALLTYPE SystemFunction_Release(ISystemFunction *this)
{
  if (--((IRealSystemFunction *)this)->dwCount == 0)
  {
    GlobalFree(this);
    InterlockedDecrement(&g_nObjs);
    return 0;
  }
  return ((IRealSystemFunction *)this)->dwCount;
}


//// IDispatch

HRESULT SystemFunction_LoadRegTypeInfo(void)
{
  LPTYPELIB pTypeLib;
  HRESULT hr;

  if ((hr=LoadRegTypeLib(&CLSID_TypeLib, 1, 0, 0, &pTypeLib)) == S_OK)
  {
    if ((hr=pTypeLib->lpVtbl->GetTypeInfoOfGuid(pTypeLib, &IID_ISystemFunction, &g_SystemFunctionTypeInfo)) == S_OK)
    {
      pTypeLib->lpVtbl->Release(pTypeLib);
      g_SystemFunctionTypeInfo->lpVtbl->AddRef(g_SystemFunctionTypeInfo);
    }
  }
  return hr;
}

ULONG STDMETHODCALLTYPE SystemFunction_GetTypeInfoCount(ISystemFunction *this, UINT *pCount)
{
  *pCount=1;
  return S_OK;
}

ULONG STDMETHODCALLTYPE SystemFunction_GetTypeInfo(ISystemFunction *this, UINT itinfo, LCID lcid, ITypeInfo **pTypeInfo)
{
  HRESULT hr;

  *pTypeInfo=NULL;

  if (itinfo)
  {
    hr=ResultFromScode(DISP_E_BADINDEX);
  }
  else if (g_SystemFunctionTypeInfo)
  {
    g_SystemFunctionTypeInfo->lpVtbl->AddRef(g_SystemFunctionTypeInfo);
    hr=S_OK;
  }
  else
  {
    hr=SystemFunction_LoadRegTypeInfo();
  }
  if (hr == S_OK) *pTypeInfo=g_SystemFunctionTypeInfo;

  return hr;
}

ULONG STDMETHODCALLTYPE SystemFunction_GetIDsOfNames(ISystemFunction *this, REFIID riid, LPOLESTR *rgszNames, UINT cNames, LCID lcid, DISPID *rgdispid)
{
  if (!g_SystemFunctionTypeInfo)
  {
    HRESULT hr;

    if ((hr=SystemFunction_LoadRegTypeInfo()) != S_OK) return hr;
  }
  return DispGetIDsOfNames(g_SystemFunctionTypeInfo, rgszNames, cNames, rgdispid);
}

ULONG STDMETHODCALLTYPE SystemFunction_Invoke(ISystemFunction *this, DISPID dispid, REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS *params, VARIANT *result, EXCEPINFO *pexcepinfo, UINT *puArgErr)
{
  if (!IsEqualIID(riid, &IID_NULL))
    return DISP_E_UNKNOWNINTERFACE;

  if (!g_SystemFunctionTypeInfo)
  {
    HRESULT hr;

    if ((hr=SystemFunction_LoadRegTypeInfo()) != S_OK) return hr;
  }
  return DispInvoke(this, g_SystemFunctionTypeInfo, dispid, wFlags, params, result, pexcepinfo, puArgErr);
}


//// ISystemFunction methods

int __declspec(naked) CallProc(HSTACK *hStack, FARPROC lpProcedure)
{
  PARAMETER *lpElement;
  DWORD dwValue;
  int nProcedure;
  int nResult;

  __asm
  {
    //Prolog
    push  ebp
    mov   ebp, esp
    sub   esp, __LOCAL_SIZE
  }

  //Push parameters to stack
  lpElement=(PARAMETER *)hStack->first;
  nProcedure=(int)lpProcedure;

  while (lpElement)
  {
    dwValue=lpElement->dwValue;

    __asm
    {
      push  [dwValue]
    }

    lpElement=lpElement->next;
  }

  //Free call parameters
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);

  __asm
  {
    //Call
    call  [nProcedure]
    mov   [nResult], eax
  }

  __asm
  {
    //Return
    mov   eax, [nResult]

    //Epilog
    mov  esp, ebp
    pop  ebp
    ret
  }
}

HRESULT STDMETHODCALLTYPE SystemFunction_AddParameter(ISystemFunction *this, VARIANT vtParameter)
{
  HSTACK *hStack;
  PARAMETER *lpElement;
  DWORD dwParameter=0;

  hStack=&((IRealSystemFunction *)this)->sf.hParametersStack;

  if (vtParameter.vt == VT_BOOL)
    dwParameter=vtParameter.boolVal;
  else if (vtParameter.vt == VT_I2)
    dwParameter=vtParameter.iVal;
  else if (vtParameter.vt == VT_I4)
    dwParameter=vtParameter.lVal;
  else if (vtParameter.vt == VT_R8)
    dwParameter=(DWORD)vtParameter.dblVal;
  else if (vtParameter.vt == VT_DISPATCH)
    dwParameter=(DWORD)vtParameter.pdispVal;
  else if (vtParameter.vt == VT_BSTR)
    dwParameter=(DWORD)xatoiW(vtParameter.bstrVal);

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, 1, sizeof(PARAMETER)))
  {
    lpElement->dwValue=dwParameter;
  }
  else return E_OUTOFMEMORY;

  return NOERROR;
}

HRESULT STDMETHODCALLTYPE SystemFunction_Call(ISystemFunction *this, BSTR wpDllFunction, SAFEARRAY **psa, int *nResult)
{
  SYSTEMFUNCTION *sf;
  HMODULE hModule;
  FARPROC lpProcedure;
  char szDll[MAX_PATH]="";
  wchar_t wszDll[MAX_PATH]=L"";
  char szFunction[MAX_PATH]="";
  wchar_t wszFunction[MAX_PATH]=L"";
  BOOL bLoadLibrary=FALSE;
  int i;

  //Fill call stack
  {
    VARIANT *vtParameter;
    unsigned char *lpData;
    DWORD dwElement;
    DWORD dwElementSum;
  
    lpData=(unsigned char *)((*psa)->pvData);
    dwElementSum=(*psa)->rgsabound[0].cElements;
  
    if (dwElementSum)
    {
      for (dwElement=0; dwElement < dwElementSum; ++dwElement)
      {
        vtParameter=(VARIANT *)(lpData + dwElement * sizeof(VARIANT));
        SystemFunction_AddParameter(this, *vtParameter);
      }
    }
  }

  *nResult=0;
  sf=&((IRealSystemFunction *)this)->sf;

  if (*wpDllFunction)
  {
    if (bOldWindows)
    {
      char szDllFunction[MAX_PATH];

      WideCharToMultiByte(CP_ACP, 0, wpDllFunction, -1, szDllFunction, MAX_PATH, NULL, NULL);

      for (i=0; szDllFunction[i]; ++i)
      {
        if (szDllFunction[i] == ':' && szDllFunction[i + 1] == ':')
        {
          lstrcpynA(szDll, szDllFunction, i + 1);
          lstrcpynA(szFunction, szDllFunction + i + 2, MAX_PATH);
          break;
        }
      }

      if (!(hModule=GetModuleHandleA(szDll)))
      {
        if (hModule=LoadLibraryA(szDll))
        {
          bLoadLibrary=TRUE;
        }
      }
    }
    else
    {
      for (i=0; wpDllFunction[i]; ++i)
      {
        if (wpDllFunction[i] == ':' && wpDllFunction[i + 1] == ':')
        {
          lstrcpynW(wszDll, wpDllFunction, i + 1);
          lstrcpynW(wszFunction, wpDllFunction + i + 2, MAX_PATH);
          break;
        }
      }

      if (!(hModule=GetModuleHandleW(wszDll)))
      {
        if (hModule=LoadLibraryW(wszDll))
        {
          bLoadLibrary=TRUE;
        }
      }
      WideCharToMultiByte(CP_ACP, 0, wszFunction, -1, szFunction, MAX_PATH, NULL, NULL);
    }

    if (hModule)
    {
      if (lpProcedure=GetProcAddress(hModule, szFunction))
      {
        //Call function
        *nResult=CallProc(&sf->hParametersStack, lpProcedure);

        //Get last error
        sf->dwLastError=GetLastError();
      }
      if (bLoadLibrary)
      {
        FreeLibrary(hModule);
      }
    }
  }
  return NOERROR;
}

HRESULT STDMETHODCALLTYPE SystemFunction_GetLastError(ISystemFunction *this, DWORD *dwLastError)
{
  *dwLastError=((IRealSystemFunction *)this)->sf.dwLastError;

  return NOERROR;
}
